var Shaders = {};
Shaders["blit-encode-float-rgba-frag.shader"] = `#include "simple-common.shader"

varying vec2 v_TexCoord;
varying vec4 v_Position;

// the texture
uniform sampler2D u_Texture;
uniform float u_Factor;

// Reference:
// * WebGL GPGPU experiment - reading a floating point texture
//   http://lab.concord.org/experiments/webgl-gpgpu/webgl.html
// * Stackoverflow - WebGL Read pixels from floating point render target
//   https://stackoverflow.com/questions/17981163/webgl-read-pixels-from-floating-point-render-target

float shift_right (float v, float amt) {
    v = floor(v) + 0.5;
    return floor(v / exp2(amt));
}
float shift_left (float v, float amt) {
    return floor(v * exp2(amt) + 0.5);
}
float mask_last (float v, float bits) {
    return mod(v, shift_left(1.0, bits));
}
float extract_bits (float num, float from, float to) {
    from = floor(from + 0.5); to = floor(to + 0.5);
    return mask_last(shift_right(num, from), to - from);
}
vec4 encode_float (float val) {
    if (val == 0.0) return vec4(0, 0, 0, 0);
    float sign = val > 0.0 ? 0.0 : 1.0;
    val = abs(val);
    float exponent = floor(log2(val));
    float biased_exponent = exponent + 127.0;
    float fraction = ((val / exp2(exponent)) - 1.0) * 8388608.0;
    float t = biased_exponent / 2.0;
    float last_bit_of_biased_exponent = fract(t) * 2.0;
    float remaining_bits_of_biased_exponent = floor(t);
    float byte4 = extract_bits(fraction, 0.0, 8.0) / 255.0;
    float byte3 = extract_bits(fraction, 8.0, 16.0) / 255.0;
    float byte2 = (last_bit_of_biased_exponent * 128.0 + extract_bits(fraction, 16.0, 23.0)) / 255.0;
    float byte1 = (sign * 128.0 + remaining_bits_of_biased_exponent) / 255.0;
    return vec4(byte4, byte3, byte2, byte1);
}

void main() {
    vec3 color = texture2D(u_Texture, v_TexCoord).rgb * u_Factor;
    // NOTE: can only encode one 32-bit float to RGBA8. Only encoding R channel here...
    gl_FragColor = encode_float(color.r);
}
`;

Shaders["blit-frag.shader"] = `#include "simple-common.shader"

varying vec2 v_TexCoord;
varying vec4 v_Position;

// the texture
uniform sampler2D u_Texture;
uniform float u_Factor;


void main()
{
    vec3 color = texture2D(u_Texture, v_TexCoord).rgb * u_Factor;
#ifdef ENABLE_GAMMA_CORRECTION
    color = pow(color, vec3(1.0 / 2.2));
#endif
    gl_FragColor = vec4(color, 1.0);
}
`;

Shaders["blit-running-mean-frag.shader"] = `#include "simple-common.shader"

varying vec2 v_TexCoord;
varying vec4 v_Position;

uniform sampler2D u_OldAvg;
uniform sampler2D u_NewSum;

uniform float u_TotalSampleCount;
uniform float u_NewSampleCount;

void main()
{
    vec3 oldAvg = texture2D(u_OldAvg, v_TexCoord).rgb;
    vec3 newSum = texture2D(u_NewSum, v_TexCoord).rgb;

    vec3 newAvg = oldAvg + (newSum - u_NewSampleCount * oldAvg) / u_TotalSampleCount;

    gl_FragColor = vec4(newAvg, 1.0);
}
`;

Shaders["blit-subtract-frag.shader"] = `#include "simple-common.shader"

varying vec2 v_TexCoord;
varying vec4 v_Position;

// subtract u_Texture1 from u_Texture0
// (result = u_Texture0 - u_Texture1)
uniform sampler2D u_Texture0;
uniform sampler2D u_Texture1;

void main()
{
    vec3 texel0 = texture2D(u_Texture0, v_TexCoord).rgb;
    vec3 texel1 = texture2D(u_Texture1, v_TexCoord).rgb;
    vec3 result = texel0 - texel1;

    gl_FragColor = vec4(result, 1.0);
}
`;

Shaders["blit-vert.shader"] = `#include "simple-common.shader"

// atrribute will receive data from a buffer
attribute vec4 a_Position;
attribute vec2 a_TexCoord;

varying vec4 v_Position;
varying vec2 v_TexCoord;

void main()
{
    gl_Position = a_Position;
    v_Position = a_Position;
    v_TexCoord = a_TexCoord;
}
`;

Shaders["ellipsoid-frag.shader"] = `#include "simple-common.shader"
#include "photon-common.shader"

in vec3 v_RayDir;
in vec3 v_FocalPos;
in float v_MinTau;
in float v_MaxTau;
in vec3 v_Color;

uniform float u_Time;
uniform vec3 u_SigmaT;
uniform vec3 u_SigmaS;
uniform bool u_IsTimeWidthZero;

out vec4 f_Color;

//
// Hash functions adapted from Nimitz:
// https://www.shadertoy.com/view/Xt3cDn
//
float g_seed = 0.;

uint base_hash(uvec2 p) {
    p = 1103515245U*((p >> 1U)^(p.yx));
    uint h32 = 1103515245U*((p.x)^(p.y>>3U));
    return h32^(h32 >> 16);
}
void init_rand(in vec2 frag_coord, in float time) {
    g_seed = float(base_hash(floatBitsToUint(frag_coord)))/float(0xffffffffU)+time;
}
float rand1(inout float seed) {
    uint n = base_hash(floatBitsToUint(vec2(seed+=.1,seed+=.1)));
    return float(n)/float(0xffffffffU);
}
vec2 rand2(inout float seed) {
    uint n = base_hash(floatBitsToUint(vec2(seed+=.1,seed+=.1)));
    uvec2 rz = uvec2(n, n*48271U);
    return vec2(rz.xy & uvec2(0x7fffffffU))/float(0x7fffffff);
}
vec3 rand3(inout float seed) {
    uint n = base_hash(floatBitsToUint(vec2(seed+=.1,seed+=.1)));
    uvec3 rz = uvec3(n, n*16807U, n*48271U);
    return vec3(rz & uvec3(0x7fffffffU))/float(0x7fffffff);
}

// NOTE: assumes origin point o is a focal point and d is normalized
bool focalRayEllipsoidIntersect(vec3 o, vec3 d, vec3 f2, float tau, out float t)
{
    // rings
    // if (abs(dot(d, o - f2) + tau) < 0.001)
    // {
    //     t = 0.0;
    //     return false;
    // }
    if (abs(dot(o - f2, o - f2) - tau * tau) < 0.001)
    {
        t = 0.0;
        return true;
    }

    float nominator = tau * tau - dot(o - f2, o - f2);
    float denominator = 2.0 * ( dot(d, o - f2) + tau );
    t = nominator / denominator;
    if (t < 0.0 || t > tau)
    {
        return false;
    }
    return true;
}

void main()
{
    vec3 rayDir = normalize(v_RayDir);

    init_rand(rayDir.xy, u_Time);

    float tau = u_IsTimeWidthZero ?
          v_MinTau
        : mix(v_MinTau, v_MaxTau, rand1(g_seed));

    float s1 = 0.0;
    bool hit = focalRayEllipsoidIntersect(vec3(0.0), rayDir, v_FocalPos, tau, s1);
    if (hit)
    {
        float pdf = u_IsTimeWidthZero ? 1.0 : 1.0 / (v_MaxTau - v_MinTau);
        float t1 = length(rayDir * s1 - v_FocalPos);

        vec3 estimate = v_Color * (1.0 / (4.0 * PI)) * exp(-u_SigmaT * t1) * u_SigmaS * (1.0 / (4.0 * PI)) * exp(-u_SigmaT * s1) / pdf;

        vec3 w1 = normalize(rayDir * s1 - v_FocalPos);
        vec3 psi1 = rayDir;
        float jacobian = abs(t1 * t1 * dot(w1, w1 + psi1));
        estimate /= jacobian;

        f_Color = vec4(estimate, 1.0);
    }
    else
    {
        f_Color = vec4(0., 1., 1., 0.);
    }
}
`;

Shaders["ellipsoid-vert.shader"] = `#include "simple-common.shader"

// ellipsoid primitive only support transient mode
#define ENABLE_TRANSIENT 1

in vec2 a_Position;      // positions of a quad in [-1, 1]^2
in vec3 a_FocalPos;
in float a_MinTau;
in float a_MaxTau;
in vec3 a_Color;

uniform vec3 u_FrustumCornerBL; // bottom-left
uniform vec3 u_FrustumCornerBR; // bottom-right
uniform vec3 u_FrustumCornerTL; // top-left
uniform vec3 u_FrustumCornerTR; // top-right

out vec3 v_RayDir;
out vec3 v_FocalPos;
out float v_MinTau;
out float v_MaxTau;
out vec3 v_Color;

void main()
{
    gl_Position = vec4(a_Position, 0.0, 1.0);
    vec2 uv = (a_Position + vec2(1.0)) / 2.0;

    float u = uv.x;
    float v = uv.y;
    vec3 rayDir = (1.0 - u) * (1.0 - v) * u_FrustumCornerBL
                + u * (1.0 - v) * u_FrustumCornerBR
                + v * (1.0 - u) * u_FrustumCornerTL
                + u * v * u_FrustumCornerTR
                ;
    v_RayDir = normalize(rayDir);
    v_FocalPos = a_FocalPos;
    v_MinTau = a_MinTau;
    v_MaxTau = a_MaxTau;
    v_Color = a_Color;
}
`;

Shaders["photon-common.shader"] = `#include "simple-common.shader"

// NOTE: keep same as SearchLightPhotonTracer.prototype.evalPhase()
float evalPhase(float _g, vec3 wi, vec3 wo)
{
    float cosTheta = dot(wi, wo);
    float term = 1.0 + _g*_g - 2.0*_g*cosTheta;
    return (1.0 / (4.0 * PI)) * (1.0 - _g * _g) / (term * sqrt(term));
}
`;

Shaders["plane-frag.shader"] = `#include "simple-common.shader"
#include "photon-common.shader"

varying vec4 v_Position;
varying vec4 v_ViewPosition;
varying vec3 v_Color;
varying vec3 v_ViewPhotonDirection;
varying vec3 v_ViewPlaneNormal;

#ifdef ENABLE_TRANSIENT
varying vec2 v_PlanePos;
varying float v_InitDist;

// u_DistInterval.x: minimum light ray path length for path to fall within time interval
// u_DistInterval.y: maximum light ray ...
uniform vec2 u_DistInterval;
#endif
// u_EvalCamTransmittance: if the shader need to calculate camera tranmittance and
//                         multiply it to color
uniform bool u_EvalCamTransmittance;
uniform vec3 u_SigmaT;
uniform bool u_IsPerspective;

// u_CameraWarped: if set to true, take camera time into account when rendering
//                 transient state
uniform bool u_CameraWarped;

void main()
{
    vec3 color = v_Color;

    float camDist = 0.0;
    if (u_EvalCamTransmittance)
    {
        float t = u_IsPerspective ? length(v_ViewPosition.xyz) : abs(v_ViewPosition.z);
        if (u_CameraWarped)
        {
            camDist = t;
        }
        vec3 transmittance = exp(-u_SigmaT*t);
        color *= transmittance;
    }

    if (u_IsPerspective)
    {
        // calculate phase funciton of w_c and w_b
        vec3 w_c = -normalize(v_ViewPosition.xyz);
        vec3 w_b = normalize(v_ViewPhotonDirection);
        // NOTE: assume g = 0.0
        float phase = evalPhase(0.0, w_c, w_b);
        color *= vec3(phase);

        // calculate jacobian term
        float jacobian = abs(dot(-w_c, v_ViewPlaneNormal));
        if (jacobian < 0.01)
        {
            color = vec3(0.0);
        }
        else
        {
            color /= jacobian;
        }
    }

#ifdef ENABLE_TRANSIENT
    float dist = v_InitDist + v_PlanePos.x + v_PlanePos.y + camDist;

    float distFactor = 0.0;
    if(dist > u_DistInterval.x && dist < u_DistInterval.y)
    {
        distFactor = 1.0;
    }
    color *= distFactor;
#endif

    gl_FragColor = vec4(color, 1.0);
}
`;

Shaders["plane-vert.shader"] = `#include "simple-common.shader"

// atrribute will receive data from a buffer
attribute vec3 a_Position;
attribute mat3 a_Transform;
attribute vec3 a_Offset;

attribute vec3 a_Color;
attribute vec2 a_PlanePos;
attribute float a_InitDist;
attribute vec3 a_PhotonDirection;
attribute vec3 a_PlaneNormal;

uniform mat4 u_ModelViewProjection;
uniform mat4 u_ModelView;

varying vec4 v_Position;
varying vec4 v_ViewPosition;
varying vec3 v_Color;
varying vec3 v_ViewPhotonDirection;
varying vec3 v_ViewPlaneNormal;
#ifdef ENABLE_TRANSIENT
varying vec2 v_PlanePos;
varying float v_InitDist;
#endif

void main()
{
    vec3 worldPos = a_Transform * a_Position + a_Offset;
    gl_Position = u_ModelViewProjection * vec4(worldPos, 1.0);
    v_Position = vec4(worldPos, 1.0);
    v_ViewPosition = u_ModelView * vec4(worldPos, 1.0);
    v_Color = a_Color;
    v_ViewPhotonDirection = (u_ModelView * vec4(a_PhotonDirection, 0.0)).xyz;
    v_ViewPlaneNormal = (u_ModelView * vec4(a_PlaneNormal, 0.0)).xyz;

#ifdef ENABLE_TRANSIENT
    // a_Position is normalized in [0, 1]^2
    v_PlanePos = a_PlanePos * a_Position.xy;
    v_InitDist = a_InitDist;
#endif
}
`;

Shaders["simple-common.shader"] = `precision highp float;

const float PI = 3.1415926535897932384626433832795;
const float PI_2 = 1.57079632679489661923;
const float PI_4 = 0.785398163397448309616;

float det(mat3 m)
{
  float a00 = m[0][0], a01 = m[0][1], a02 = m[0][2];
  float a10 = m[1][0], a11 = m[1][1], a12 = m[1][2];
  float a20 = m[2][0], a21 = m[2][1], a22 = m[2][2];
  float b01 = a22 * a11 - a12 * a21;
  float b11 = -a22 * a10 + a12 * a20;
  float b21 = a21 * a10 - a11 * a20;
  float det = a00 * b01 + a01 * b11 + a02 * b21;
  return det;
}

#ifndef USE_GLES30
// inverse a matrix in GLSL
// copied from: https://github.com/glslify/glsl-inverse/blob/master/index.glsl
float inverse(float m) {
  return 1.0 / m;
}

mat2 inverse(mat2 m) {
  return mat2(m[1][1],-m[0][1],
             -m[1][0], m[0][0]) / (m[0][0]*m[1][1] - m[0][1]*m[1][0]);
}

mat3 inverse(mat3 m) {
  float a00 = m[0][0], a01 = m[0][1], a02 = m[0][2];
  float a10 = m[1][0], a11 = m[1][1], a12 = m[1][2];
  float a20 = m[2][0], a21 = m[2][1], a22 = m[2][2];

  float b01 = a22 * a11 - a12 * a21;
  float b11 = -a22 * a10 + a12 * a20;
  float b21 = a21 * a10 - a11 * a20;

  float det = a00 * b01 + a01 * b11 + a02 * b21;

  return mat3(b01, (-a22 * a01 + a02 * a21), (a12 * a01 - a02 * a11),
              b11, (a22 * a00 - a02 * a20), (-a12 * a00 + a02 * a10),
              b21, (-a21 * a00 + a01 * a20), (a11 * a00 - a01 * a10)) / det;
}

mat4 inverse(mat4 m) {
  float
      a00 = m[0][0], a01 = m[0][1], a02 = m[0][2], a03 = m[0][3],
      a10 = m[1][0], a11 = m[1][1], a12 = m[1][2], a13 = m[1][3],
      a20 = m[2][0], a21 = m[2][1], a22 = m[2][2], a23 = m[2][3],
      a30 = m[3][0], a31 = m[3][1], a32 = m[3][2], a33 = m[3][3],

      b00 = a00 * a11 - a01 * a10,
      b01 = a00 * a12 - a02 * a10,
      b02 = a00 * a13 - a03 * a10,
      b03 = a01 * a12 - a02 * a11,
      b04 = a01 * a13 - a03 * a11,
      b05 = a02 * a13 - a03 * a12,
      b06 = a20 * a31 - a21 * a30,
      b07 = a20 * a32 - a22 * a30,
      b08 = a20 * a33 - a23 * a30,
      b09 = a21 * a32 - a22 * a31,
      b10 = a21 * a33 - a23 * a31,
      b11 = a22 * a33 - a23 * a32,

      det = b00 * b11 - b01 * b10 + b02 * b09 + b03 * b08 - b04 * b07 + b05 * b06;

  return mat4(
      a11 * b11 - a12 * b10 + a13 * b09,
      a02 * b10 - a01 * b11 - a03 * b09,
      a31 * b05 - a32 * b04 + a33 * b03,
      a22 * b04 - a21 * b05 - a23 * b03,
      a12 * b08 - a10 * b11 - a13 * b07,
      a00 * b11 - a02 * b08 + a03 * b07,
      a32 * b02 - a30 * b05 - a33 * b01,
      a20 * b05 - a22 * b02 + a23 * b01,
      a10 * b10 - a11 * b08 + a13 * b06,
      a01 * b08 - a00 * b10 - a03 * b06,
      a30 * b04 - a31 * b02 + a33 * b00,
      a21 * b02 - a20 * b04 - a23 * b00,
      a11 * b07 - a10 * b09 - a12 * b06,
      a00 * b09 - a01 * b07 + a02 * b06,
      a31 * b01 - a30 * b03 - a32 * b00,
      a20 * b03 - a21 * b01 + a22 * b00) / det;
}
#endif

void swap(inout float a, inout float b)
{
    float temp = a;
    a = b;
    b = temp;
}



`;

Shaders["simple-frag.shader"] = `#include "simple-common.shader"

void main()
{
    gl_FragColor = vec4(1, 1, 1, 1);  // return reddish-purple
}
`;

Shaders["simple-vert.shader"] = `#include "simple-common.shader"

// atrribute will receive data from a buffer
attribute vec4 a_Position;
uniform mat4 u_ModelViewProjection;

void main()
{
    gl_Position = u_ModelViewProjection * a_Position;
}
`;

Shaders["sphere-frag.shader"] = `#include "simple-common.shader"
#include "photon-common.shader"

in vec3 v_ViewPosition;
in vec3 v_ViewSphereCenter;
in vec3 v_Color;
in vec3 v_InDirection;
in float v_Radius;

out vec4 f_Color;

uniform vec2 u_Resolution;
uniform mat4 u_InvProjection;
uniform vec3 u_SigmaT;

bool raySphereIntersect(vec3 o, vec3 d, vec3 c, float r, out float t1, out float t2)
{
    vec3 l = c - o;
    float s = dot(l, d);
    float lsqr = dot(l, l);
    float rsqr = r * r;
    if (s < 0.0 && lsqr > rsqr)
    {
        return false;
    }

    float msqr = lsqr - s * s;
    if (msqr > rsqr)
    {
        return false;
    }

    float q = sqrt(rsqr - msqr);
    t1 = s - q;
    t2 = s + q;
    return true;
}

void main()
{
    // assume orthographic camera
    vec2 uv = gl_FragCoord.xy / u_Resolution.xy;
    vec3 ndc = vec3(uv * 2.0 - 1.0, -1.0);
    vec3 camRayOrigin = (u_InvProjection * vec4(ndc, 1.0)).xyz;
    vec3 camRayDir = vec3(0.0, 0.0, -1.0);

    f_Color = vec4(0.0,0.0,0.0,1.0);
    float t1 = 0.0;
    float t2 = 0.0;
    if (!raySphereIntersect(camRayOrigin, camRayDir, v_ViewSphereCenter, v_Radius, t1, t2))
    {
        return;
    }

    vec2 ts = vec2(t1, t2);
    for (int i = 0; i < 2; ++i)
    {
        float t = ts[i];
        if (t < 0.0)
        {
            continue;
        }

        vec3 hitPoint = camRayOrigin + camRayDir * t;
        vec3 offset = hitPoint - v_ViewSphereCenter;
        vec3 w1 = normalize(offset);
        float t1 = v_Radius;

        if (dot(w1, w1) == 0.f || t1 == 0.f)
        {
            continue;
        }

        vec3 n = t1 * t1 * w1;
        float jacobian = abs(dot(n, camRayDir));
        // if (jacobian < 1e-5f)
        // {
        //     continue;
        // }
        float invJacobian = 1.0 / jacobian;

        vec3 camTrans = exp(-u_SigmaT * t);

        // NOTE: assume g == 0.0
        float lastScatter = evalPhase(0.0, v_InDirection, w1);
        float fp1 = evalPhase(0.0, w1, -camRayDir);
        vec3 estimate = lastScatter * fp1 * u_SigmaT * v_Color * invJacobian * camTrans;

        // NOTE: WARNING: code still produces nan/inf
        // TODO: figure out why
        if (any(isnan(estimate)) || any(isinf(estimate)))
        {
            continue;
        }

        f_Color.xyz += estimate;
    }
}
`;

Shaders["sphere-vert.shader"] = `#include "simple-common.shader"

// sphere primitive only support transient mode
#define ENABLE_TRANSIENT 1

in vec3 a_Position;     // vertices of cube of [-1,1]^3
in vec3 a_Offset;
in float a_Radius;
in vec3 a_InDirection;
in vec3 a_Color;

uniform mat4 u_ModelViewProjection;
uniform mat4 u_ModelView;

out vec3 v_ViewPosition;
out vec3 v_ViewSphereCenter;
out vec3 v_Color;
out vec3 v_InDirection;
out float v_Radius;


void main()
{
    vec3 worldPos = a_Radius * a_Position + a_Offset;
    gl_Position = u_ModelViewProjection * vec4(worldPos, 1.0);
    v_ViewPosition = (u_ModelView * vec4(worldPos, 1.0)).xyz;
    v_ViewSphereCenter = (u_ModelView * vec4(a_Offset, 1.0)).xyz;
    v_Color = a_Color;
    v_InDirection = a_InDirection;
    v_Radius = a_Radius;
}
`;

Shaders["volume-frag.shader"] = `#include "simple-common.shader"
#include "photon-common.shader"

varying vec4 v_ViewPosition;
// transform from the unit cube space, defined by the photon volume, to the camera space
varying mat4 v_UnitCubeToCamera;
varying mat4 v_CameraToUnitCube;

varying vec3 v_Color;
varying vec2 v_DistInterval;
varying vec3 v_LastPhotonDirection;

// NOTE: only works for perspective camera
// ray shooting from camera to the vertex, in camera space
// varying vec3 v_CameraRayDir;

#ifdef ENABLE_TRANSIENT
// TRANSIENT PART
// transient planes in unit cube space
// formula for such a plane: \vec{n} \cdot \vec{p - p_0} = 0
// where p is a point on plane
varying vec3 v_CubeMinPlaneNormal;  // n
varying vec3 v_CubeMinPlanePoint;   // p
varying vec3 v_CubeMaxPlaneNormal;  // n
varying vec3 v_CubeMaxPlanePoint;   // p
varying float v_DistWidth;
varying float v_CubeWorldDiagLen;
varying vec3 v_VolumeSideLengths;
#endif

uniform vec2 u_Resolution;
uniform mat4 u_InvProjection;

// u_EvalCamTransmittance: if the shader need to calculate camera tranmittance and
//                         multiply it to color
uniform bool u_EvalCamTransmittance;
uniform vec3 u_SigmaT;

// u_CameraWarped: if set to true, take camera time into account when rendering
//                 transient state
uniform bool u_CameraWarped;

const float INTERSECT_EPSILON = 1e-9;

// Intersect a ray o + td with a unit cube
// One of the corners of the unit cube is on (0, 0, 0)
// And the three corners of the unit cube aligns with the +x, +y, +z axis.
// NOTE: If the ray starts from the inside of the unit cube,
//       tMin will output as 0.
//       This behavior will make it easy for us the calculate the distance
//       a ray has traversed inside the unit cube.
bool rayUnitCubeIntersection(vec3 o, vec3 d, out float tMin, out float tMax, out vec3 reason)
{
    bool isMinSet = false;
    bool isMaxSet = false;

    for(int dim = 0; dim < 3; ++dim)
    {
        float p0 = 0.0;
        float p1 = 1.0;
        vec3 n = vec3(0.0);
        n[dim] = 1.0;

        float f = dot(d, n);

        if(abs(f) > INTERSECT_EPSILON)
        {
            float t1 = (p0 - o[dim]) / f;
            float t2 = (p1 - o[dim]) / f;

            if(t1 > t2)
            {
                swap(t1, t2);
            }

            if(!isMinSet || t1 > tMin)
            {
                tMin = t1;
                isMinSet = true;
            }
            if(!isMaxSet || t2 < tMax)
            {
                tMax = t2;
                isMaxSet = true;
            }

            if(tMin > tMax)
            {
                reason = vec3(1.0, 0.0, 0.0);
                return false;
            }
            if(tMax < 0.0)
            {
                reason = vec3(0.0, 1.0, 0.0);
                return false;
            }
        }
        else if(o[dim] < 0.0 || o[dim] > 1.0)
        {
            reason = vec3(0.0, 0.0, 1.0);
            return false;
        }
    }
    if(tMin < 0.0)
    {
        // ray origin inside unit cube?
        tMin = 0.0;
    }
    return true;
}

bool rayPlaneIntersect(vec3 o, vec3 d, vec3 p0, vec3 n, out float t)
{
    float a = dot(d, n);
    if(abs(a) < INTERSECT_EPSILON)
    {
        t = -1.0;
        return false;
    }

    t = 1.0 / a * dot((p0 - o), n);
    return true;
}

bool rangeIntersect(float aMin, float aMax, float bMin, float bMax, out float oMin, out float oMax)
{
    oMin = max(aMin, bMin);
    oMax = min(aMax, bMax);
    if(oMin > oMax)
    {
        return false;
    }
    else
    {
        return true;
    }
}

float lengthInCameraSpace(vec3 unitDir, mat4 unitCubeToCamera)
{
    vec3 camDir = (unitCubeToCamera * vec4(unitDir, 0.0)).xyz;
    float lenFactor = length(camDir);
    return lenFactor;
}

vec3 getThroughput(vec3 unitOrigin, vec3 unitDir, float tMin, float tMax, mat4 unitCubeToCamera, float phase)
{
    // vec4 p0 = vec4(unitOrigin, 1.0) + vec4(unitDir, 0.0) * tMin;
    // vec4 p1 = vec4(unitOrigin, 1.0) + vec4(unitDir, 0.0) * tMax;

    // p0 = unitCubeToCamera * p0;
    // p1 = unitCubeToCamera * p1;

    // float thickness = length(p1 - p0);
    // return thickness;
    float lenFactor = lengthInCameraSpace(unitDir, unitCubeToCamera);
    vec3 ret = vec3(phase);

    if(!u_EvalCamTransmittance)
    {
        // just need to calculate the distance between the two intersection points in cam/world space
        // i.e. the thickness of the intersection of camera ray and photon volume
        ret *= vec3(abs(tMax - tMin) * lenFactor);
    }
    else
    {
        tMin = abs(tMin);
        tMax = abs(tMax);
        if(tMin > tMax)
        {
            float temp = tMin;
            tMin = tMax;
            tMax = temp;
        }

        // tMin and tMax in camera space
        float tMinCam = tMin * lenFactor;
        float tMaxCam = tMax * lenFactor;

        ret *= -1.0 / u_SigmaT * (exp(-u_SigmaT * tMaxCam) - exp(-u_SigmaT * tMinCam));
    }

    return ret;
}

void main()
{
    vec2 uv = gl_FragCoord.xy / u_Resolution.xy;
    vec3 ndc = vec3(uv * 2.0 - 1.0, -1.0);

    // ray origin, dir in camera (view) space
    vec4 camRayOrigin = vec4(0.0, 0.0, 0.0, 1.0);
    vec4 camRayDir = vec4(0.0);
    if (u_CameraWarped)
    {
        camRayDir = vec4(normalize(v_ViewPosition.xyz), 0.0);
    }
    else
    {
        camRayOrigin = (u_InvProjection * vec4(ndc, 1.0));
        camRayOrigin.w = 1.0;
        camRayDir = vec4(0, 0, -1, 0);
    }

    // NOTE: assume g == 0
    float phase = evalPhase(0.0, -camRayDir.xyz, v_LastPhotonDirection);


    // camera space -> unit cube space
    vec4 unitCubeRayOrigin = v_CameraToUnitCube * camRayOrigin;
    vec4 unitCubeRayDir = normalize(v_CameraToUnitCube * camRayDir);

    float tMin = -1.0;
    float tMax = -1.0;
    vec3 reason = vec3(0.0);

    vec3 origin = unitCubeRayOrigin.xyz;
    vec3 dir = unitCubeRayDir.xyz;

    if(rayUnitCubeIntersection(origin, dir, tMin, tMax, reason))
    {
#ifdef ENABLE_TRANSIENT
        // TRANSIENT part code
        if (u_CameraWarped)
        {
            vec3 minUvw = origin + dir * tMin;
            vec3 maxUvw = origin + dir * tMax;
            float minVolumeLen = dot(minUvw, v_VolumeSideLengths);
            float maxVolumeLen = dot(maxUvw, v_VolumeSideLengths);

            float lenFactor = lengthInCameraSpace(dir, v_UnitCubeToCamera);
            float minCamLen = tMin * lenFactor;
            float maxCamLen = tMax * lenFactor;

            float minTotalLen = minCamLen + minVolumeLen;
            float maxTotalLen = maxCamLen + maxVolumeLen;

            if (minTotalLen > maxTotalLen)
            {
                swap(minTotalLen, maxTotalLen);
                swap(tMin, tMax);
            }

            float oMin = -1.0;
            float oMax = -1.0;

            gl_FragColor = vec4(0.0);
            if(rangeIntersect(minTotalLen, maxTotalLen, v_DistInterval.x, v_DistInterval.y, oMin, oMax))
            {
                float q1 = (oMin - minTotalLen) / (maxTotalLen - minTotalLen);
                float q2 = (oMax - minTotalLen) / (maxTotalLen - minTotalLen);

                q1 = clamp(q1, 0.0, 1.0);
                q2 = clamp(q2, 0.0, 1.0);

                float t1 = tMin * (1.0 - q1) + tMax * q1;
                float t2 = tMin * (1.0 - q2) + tMax * q2;

                if (t1 > t2)
                {
                    swap(t1, t2);
                }

                vec3 throughput = getThroughput(origin, dir, t1, t2, v_UnitCubeToCamera, phase);
                gl_FragColor = vec4(throughput * v_Color, 1.0);
            }
            return;
        }

        if(v_DistWidth > 0.0)
        {
            // still rendering "photon volume"
            float tTransMin = -1.0;
            float tTransMax = -1.0;

            bool hitMin = rayPlaneIntersect(origin, dir, v_CubeMinPlanePoint, v_CubeMinPlaneNormal, tTransMin);
            bool hitMax = rayPlaneIntersect(origin, dir, v_CubeMaxPlanePoint, v_CubeMaxPlaneNormal, tTransMax);

            if(tTransMin > tTransMax)
            {
                swap(tTransMin, tTransMax);
            }

            if(hitMin && hitMax)
            {
                float oMin = -1.0;
                float oMax = -1.0;
                if(rangeIntersect(tMin, tMax, tTransMin, tTransMax, oMin, oMax))
                {
                    tMin = oMin;
                    tMax = oMax;

                    vec3 throughput = getThroughput(origin, dir, tMin, tMax, v_UnitCubeToCamera, phase);
                    vec3 color = v_Color;
                    color *= throughput;
                    gl_FragColor = vec4(color, 1.0);

                    // NOTE: noise is in here
                    // NOTE: only for debug
                    // gl_FragColor = vec4(1.0, 0.0, 0.0, 1.0);
                }
                else
                {
                    gl_FragColor = vec4(0.0);
                }
            }
            else if(!hitMin && !hitMax)
            {
                // ray might be inside two transient planes, do a distance test
                float d0 = dot(origin - v_CubeMinPlanePoint, v_CubeMinPlaneNormal);
                float d1 = dot(origin - v_CubeMaxPlanePoint, v_CubeMaxPlaneNormal);
                // assume two transient planes have same plane normal
                if(d0 * d1 < 0.0)
                {
                    // different sign of d0 and d1 means the ray is inside two transient planes
                    vec3 throughput = getThroughput(origin, dir, tMin, tMax, v_UnitCubeToCamera, phase);
                    vec3 color = v_Color;
                    color *= throughput;
                    gl_FragColor = vec4(color, 1.0);

                    // NOTE: only for debug:
                    // gl_FragColor = vec4(0.0, 1.0, 0.0, 0.0);
                }
                else
                {
                    gl_FragColor = vec4(0.0);
                    // NOTE: only for debug:
                    // gl_FragColor = vec4(0.0, 1.0, 1.0, 0.0);
                }
            }
            else
            {
                // shouldn't happening, output magenta color for debug
                gl_FragColor = vec4(1.0, 0.0, 1.0, 1.0);
                return;
            }
        }
        else if(v_DistWidth == 0.0)
        {
            // rendering unbiased "transient photon plane"
            float tTrans = -1.0;
            bool hit = rayPlaneIntersect(origin, dir, v_CubeMinPlanePoint, v_CubeMinPlaneNormal, tTrans);
            vec4 outColor = vec4(0.0);
            if(hit)
            {
                if(tTrans >= tMin && tTrans <= tMax)
                {
                    // NOTE: assume speed of light is 1
                    const float speedOfLight = 1.0;
                    vec3 color = v_Color;

                    // NOTE: old way of calculating brightness of trasient plane
                    // float unitCubeSpaceFactor = length((v_UnitCubeToCamera * vec4(dir, 0.0)).xyz);

                    // float rayPlaneFactor = speedOfLight / (v_CubeWorldDiagLen * abs(dot(dir, v_CubeMinPlaneNormal)));
                    // color = unitCubeSpaceFactor * rayPlaneFactor * color;

                    vec3 e = normalize(v_UnitCubeToCamera[0].xyz);
                    vec3 f = normalize(v_UnitCubeToCamera[1].xyz);
                    vec3 g = normalize(v_UnitCubeToCamera[2].xyz);

                    vec3 nw = cross(e, f) + cross(f, g) + cross(g, e);
                    float dot_prod = abs(dot(nw, camRayDir.xyz));
                    // multiply determinant back
                    float det = abs(dot(e, cross(f, g)));
                    color = color * det / dot_prod;

                    if (u_EvalCamTransmittance)
                    {
                        // camera transmittance estimate
                        float camToHitPointLength = tTrans * lengthInCameraSpace(dir, v_UnitCubeToCamera);
                        vec3 camTrans = exp(-u_SigmaT * camToHitPointLength);
                        color *= camTrans * phase;
                    }

                    outColor = vec4(color, 1.0);
                }
            }
            gl_FragColor = outColor;
        }
        else
        {
            // shouldn't happening, output magenta color for debug
            gl_FragColor = vec4(1.0, 0.0, 1.0, 1.0);
            return;
        }
#else
        // non transient part code
        vec3 throughput = getThroughput(origin, dir, tMin, tMax, v_UnitCubeToCamera, phase);
        vec3 color = v_Color;
        color *= throughput;
        gl_FragColor = vec4(color, 1.0);
#endif
    }
    else
    {
        // disagreement between ray tracer and rasterizer (floating point error)
        // treat as no-hit
        gl_FragColor = vec4(0.0);
    }
}
`;

Shaders["volume-vert.shader"] = `#include "simple-common.shader"

// atrribute will receive data from a buffer
attribute vec4 a_Position;
attribute mat4 a_VolumeMatrix;
attribute vec3 a_Color;
// TRANSIENT PART
// a_DistInterval.x: minimum light ray path length for path to fall within time interval
// a_DistInterval.y: maximum light ray ...
attribute vec2 a_DistInterval;

uniform mat4 u_ModelViewProjection;
uniform mat4 u_ModelView;
uniform mat4 u_View;

varying vec4 v_ViewPosition;
// transform from the unit cube space, defined by the photon volume, to the camera space
varying mat4 v_UnitCubeToCamera;
varying mat4 v_CameraToUnitCube;
varying vec3 v_Color;

varying vec2 v_DistInterval;
varying vec3 v_LastPhotonDirection;

#ifdef ENABLE_TRANSIENT
// TRANSIENT PART
// transient planes in unit cube space
// formula for such a plane: \vec{n} \cdot \vec{p - p_0} = 0
// where p is a point on plane
varying vec3 v_CubeMinPlaneNormal;  // n
varying vec3 v_CubeMinPlanePoint;   // p
varying vec3 v_CubeMaxPlaneNormal;  // n
varying vec3 v_CubeMaxPlanePoint;   // p
varying float v_DistWidth;
varying float v_CubeWorldDiagLen;
varying vec3 v_VolumeSideLengths;
#endif

// NOTE: only works for perspective camera
// ray shooting from camera to the vertex, in camera space
// varying vec3 v_CameraRayDir;

// representation of plane:
// n * (p - p0) = 0
// n * p = s
// input: n, s
// onput: p0
vec3 findPointOnPlane(vec3 n, float s)
{
    vec3 absN = abs(n);
    if(absN.x > absN.y && absN.x > absN.z)
    {
        // choose x as dividend
        return vec3(s / n.x, 0, 0);
    }
    else if(absN.y > absN.z && absN.y > absN.x)
    {
        return vec3(0, s / n.y, 0);
    }
    else
    {
        return vec3(0, 0, s / n.z);
    }
}

void main()
{
    gl_Position = u_ModelViewProjection * a_Position;

    v_ViewPosition = u_ModelView * vec4(a_Position.xyz, 1.0);
    mat4 unitCubeToWorld = a_VolumeMatrix;
    v_UnitCubeToCamera = u_View * unitCubeToWorld;
    v_CameraToUnitCube = inverse(v_UnitCubeToCamera);

    v_Color = a_Color;
    v_DistInterval = a_DistInterval;
    v_LastPhotonDirection = normalize(unitCubeToWorld[2].xyz);

#ifdef ENABLE_TRANSIENT
    // three axis of unit cube space, written in world space
    vec3 p = unitCubeToWorld[0].xyz;
    vec3 q = unitCubeToWorld[1].xyz;
    vec3 r = unitCubeToWorld[2].xyz;
    // normal of the transient plane
    vec3 n = vec3(length(p), length(q), length(r));
    v_VolumeSideLengths = n;
    float length = length(n);
    v_CubeWorldDiagLen = length;

    v_CubeMinPlaneNormal = n / length;
    v_CubeMaxPlaneNormal = n / length;

    v_CubeMinPlanePoint = findPointOnPlane(n / length, a_DistInterval.x / length);
    v_CubeMaxPlanePoint = findPointOnPlane(n / length, a_DistInterval.y / length);

    v_DistWidth = a_DistInterval.y - a_DistInterval.x;
#endif

    // vertex position in camera space
    // vec3 eyeVertPos = (u_ModelView * a_Position).xyz;

    // NOTE: only works for perspective camera
    // direction == pos because camera is at (0, 0, 0) in camera space
    // v_CameraRayDir = normalize(eyeVertPos);
}
`;

