#include "simple-common.shader"
#include "photon-common.shader"

in vec3 v_RayDir;
in vec3 v_FocalPos;
in float v_MinTau;
in float v_MaxTau;
in vec3 v_Color;

uniform float u_Time;
uniform vec3 u_SigmaT;
uniform vec3 u_SigmaS;
uniform bool u_IsTimeWidthZero;

out vec4 f_Color;

//
// Hash functions adapted from Nimitz:
// https://www.shadertoy.com/view/Xt3cDn
//
float g_seed = 0.;

uint base_hash(uvec2 p) {
    p = 1103515245U*((p >> 1U)^(p.yx));
    uint h32 = 1103515245U*((p.x)^(p.y>>3U));
    return h32^(h32 >> 16);
}
void init_rand(in vec2 frag_coord, in float time) {
    g_seed = float(base_hash(floatBitsToUint(frag_coord)))/float(0xffffffffU)+time;
}
float rand1(inout float seed) {
    uint n = base_hash(floatBitsToUint(vec2(seed+=.1,seed+=.1)));
    return float(n)/float(0xffffffffU);
}
vec2 rand2(inout float seed) {
    uint n = base_hash(floatBitsToUint(vec2(seed+=.1,seed+=.1)));
    uvec2 rz = uvec2(n, n*48271U);
    return vec2(rz.xy & uvec2(0x7fffffffU))/float(0x7fffffff);
}
vec3 rand3(inout float seed) {
    uint n = base_hash(floatBitsToUint(vec2(seed+=.1,seed+=.1)));
    uvec3 rz = uvec3(n, n*16807U, n*48271U);
    return vec3(rz & uvec3(0x7fffffffU))/float(0x7fffffff);
}

// NOTE: assumes origin point o is a focal point and d is normalized
bool focalRayEllipsoidIntersect(vec3 o, vec3 d, vec3 f2, float tau, out float t)
{
    // rings
    // if (abs(dot(d, o - f2) + tau) < 0.001)
    // {
    //     t = 0.0;
    //     return false;
    // }
    if (abs(dot(o - f2, o - f2) - tau * tau) < 0.001)
    {
        t = 0.0;
        return true;
    }

    float nominator = tau * tau - dot(o - f2, o - f2);
    float denominator = 2.0 * ( dot(d, o - f2) + tau );
    t = nominator / denominator;
    if (t < 0.0 || t > tau)
    {
        return false;
    }
    return true;
}

void main()
{
    vec3 rayDir = normalize(v_RayDir);

    init_rand(rayDir.xy, u_Time);

    float tau = u_IsTimeWidthZero ?
          v_MinTau
        : mix(v_MinTau, v_MaxTau, rand1(g_seed));

    float s1 = 0.0;
    bool hit = focalRayEllipsoidIntersect(vec3(0.0), rayDir, v_FocalPos, tau, s1);
    if (hit)
    {
        float pdf = u_IsTimeWidthZero ? 1.0 : 1.0 / (v_MaxTau - v_MinTau);
        float t1 = length(rayDir * s1 - v_FocalPos);

        vec3 estimate = v_Color * (1.0 / (4.0 * PI)) * exp(-u_SigmaT * t1) * u_SigmaS * (1.0 / (4.0 * PI)) * exp(-u_SigmaT * s1) / pdf;

        vec3 w1 = normalize(rayDir * s1 - v_FocalPos);
        vec3 psi1 = rayDir;
        float jacobian = abs(t1 * t1 * dot(w1, w1 + psi1));
        estimate /= jacobian;

        f_Color = vec4(estimate, 1.0);
    }
    else
    {
        f_Color = vec4(0., 1., 1., 0.);
    }
}