#include "simple-common.shader"
#include "photon-common.shader"

varying vec4 v_ViewPosition;
// transform from the unit cube space, defined by the photon volume, to the camera space
varying mat4 v_UnitCubeToCamera;
varying mat4 v_CameraToUnitCube;

varying vec3 v_Color;
varying vec2 v_DistInterval;
varying vec3 v_LastPhotonDirection;

// NOTE: only works for perspective camera
// ray shooting from camera to the vertex, in camera space
// varying vec3 v_CameraRayDir;

#ifdef ENABLE_TRANSIENT
// TRANSIENT PART
// transient planes in unit cube space
// formula for such a plane: \vec{n} \cdot \vec{p - p_0} = 0
// where p is a point on plane
varying vec3 v_CubeMinPlaneNormal;  // n
varying vec3 v_CubeMinPlanePoint;   // p
varying vec3 v_CubeMaxPlaneNormal;  // n
varying vec3 v_CubeMaxPlanePoint;   // p
varying float v_DistWidth;
varying float v_CubeWorldDiagLen;
varying vec3 v_VolumeSideLengths;
#endif

uniform vec2 u_Resolution;
uniform mat4 u_InvProjection;

// u_EvalCamTransmittance: if the shader need to calculate camera tranmittance and
//                         multiply it to color
uniform bool u_EvalCamTransmittance;
uniform vec3 u_SigmaT;

// u_CameraWarped: if set to true, take camera time into account when rendering
//                 transient state
uniform bool u_CameraWarped;

const float INTERSECT_EPSILON = 1e-9;

// Intersect a ray o + td with a unit cube
// One of the corners of the unit cube is on (0, 0, 0)
// And the three corners of the unit cube aligns with the +x, +y, +z axis.
// NOTE: If the ray starts from the inside of the unit cube,
//       tMin will output as 0.
//       This behavior will make it easy for us the calculate the distance
//       a ray has traversed inside the unit cube.
bool rayUnitCubeIntersection(vec3 o, vec3 d, out float tMin, out float tMax, out vec3 reason)
{
    bool isMinSet = false;
    bool isMaxSet = false;

    for(int dim = 0; dim < 3; ++dim)
    {
        float p0 = 0.0;
        float p1 = 1.0;
        vec3 n = vec3(0.0);
        n[dim] = 1.0;

        float f = dot(d, n);

        if(abs(f) > INTERSECT_EPSILON)
        {
            float t1 = (p0 - o[dim]) / f;
            float t2 = (p1 - o[dim]) / f;

            if(t1 > t2)
            {
                swap(t1, t2);
            }

            if(!isMinSet || t1 > tMin)
            {
                tMin = t1;
                isMinSet = true;
            }
            if(!isMaxSet || t2 < tMax)
            {
                tMax = t2;
                isMaxSet = true;
            }

            if(tMin > tMax)
            {
                reason = vec3(1.0, 0.0, 0.0);
                return false;
            }
            if(tMax < 0.0)
            {
                reason = vec3(0.0, 1.0, 0.0);
                return false;
            }
        }
        else if(o[dim] < 0.0 || o[dim] > 1.0)
        {
            reason = vec3(0.0, 0.0, 1.0);
            return false;
        }
    }
    if(tMin < 0.0)
    {
        // ray origin inside unit cube?
        tMin = 0.0;
    }
    return true;
}

bool rayPlaneIntersect(vec3 o, vec3 d, vec3 p0, vec3 n, out float t)
{
    float a = dot(d, n);
    if(abs(a) < INTERSECT_EPSILON)
    {
        t = -1.0;
        return false;
    }

    t = 1.0 / a * dot((p0 - o), n);
    return true;
}

bool rangeIntersect(float aMin, float aMax, float bMin, float bMax, out float oMin, out float oMax)
{
    oMin = max(aMin, bMin);
    oMax = min(aMax, bMax);
    if(oMin > oMax)
    {
        return false;
    }
    else
    {
        return true;
    }
}

float lengthInCameraSpace(vec3 unitDir, mat4 unitCubeToCamera)
{
    vec3 camDir = (unitCubeToCamera * vec4(unitDir, 0.0)).xyz;
    float lenFactor = length(camDir);
    return lenFactor;
}

vec3 getThroughput(vec3 unitOrigin, vec3 unitDir, float tMin, float tMax, mat4 unitCubeToCamera, float phase)
{
    // vec4 p0 = vec4(unitOrigin, 1.0) + vec4(unitDir, 0.0) * tMin;
    // vec4 p1 = vec4(unitOrigin, 1.0) + vec4(unitDir, 0.0) * tMax;

    // p0 = unitCubeToCamera * p0;
    // p1 = unitCubeToCamera * p1;

    // float thickness = length(p1 - p0);
    // return thickness;
    float lenFactor = lengthInCameraSpace(unitDir, unitCubeToCamera);
    vec3 ret = vec3(phase);

    if(!u_EvalCamTransmittance)
    {
        // just need to calculate the distance between the two intersection points in cam/world space
        // i.e. the thickness of the intersection of camera ray and photon volume
        ret *= vec3(abs(tMax - tMin) * lenFactor);
    }
    else
    {
        tMin = abs(tMin);
        tMax = abs(tMax);
        if(tMin > tMax)
        {
            float temp = tMin;
            tMin = tMax;
            tMax = temp;
        }

        // tMin and tMax in camera space
        float tMinCam = tMin * lenFactor;
        float tMaxCam = tMax * lenFactor;

        ret *= -1.0 / u_SigmaT * (exp(-u_SigmaT * tMaxCam) - exp(-u_SigmaT * tMinCam));
    }

    return ret;
}

void main()
{
    vec2 uv = gl_FragCoord.xy / u_Resolution.xy;
    vec3 ndc = vec3(uv * 2.0 - 1.0, -1.0);

    // ray origin, dir in camera (view) space
    vec4 camRayOrigin = vec4(0.0, 0.0, 0.0, 1.0);
    vec4 camRayDir = vec4(0.0);
    if (u_CameraWarped)
    {
        camRayDir = vec4(normalize(v_ViewPosition.xyz), 0.0);
    }
    else
    {
        camRayOrigin = (u_InvProjection * vec4(ndc, 1.0));
        camRayOrigin.w = 1.0;
        camRayDir = vec4(0, 0, -1, 0);
    }

    // NOTE: assume g == 0
    float phase = evalPhase(0.0, -camRayDir.xyz, v_LastPhotonDirection);


    // camera space -> unit cube space
    vec4 unitCubeRayOrigin = v_CameraToUnitCube * camRayOrigin;
    vec4 unitCubeRayDir = normalize(v_CameraToUnitCube * camRayDir);

    float tMin = -1.0;
    float tMax = -1.0;
    vec3 reason = vec3(0.0);

    vec3 origin = unitCubeRayOrigin.xyz;
    vec3 dir = unitCubeRayDir.xyz;

    if(rayUnitCubeIntersection(origin, dir, tMin, tMax, reason))
    {
#ifdef ENABLE_TRANSIENT
        // TRANSIENT part code
        if (u_CameraWarped)
        {
            vec3 minUvw = origin + dir * tMin;
            vec3 maxUvw = origin + dir * tMax;
            float minVolumeLen = dot(minUvw, v_VolumeSideLengths);
            float maxVolumeLen = dot(maxUvw, v_VolumeSideLengths);

            float lenFactor = lengthInCameraSpace(dir, v_UnitCubeToCamera);
            float minCamLen = tMin * lenFactor;
            float maxCamLen = tMax * lenFactor;

            float minTotalLen = minCamLen + minVolumeLen;
            float maxTotalLen = maxCamLen + maxVolumeLen;

            if (minTotalLen > maxTotalLen)
            {
                swap(minTotalLen, maxTotalLen);
                swap(tMin, tMax);
            }

            float oMin = -1.0;
            float oMax = -1.0;

            gl_FragColor = vec4(0.0);
            if(rangeIntersect(minTotalLen, maxTotalLen, v_DistInterval.x, v_DistInterval.y, oMin, oMax))
            {
                float q1 = (oMin - minTotalLen) / (maxTotalLen - minTotalLen);
                float q2 = (oMax - minTotalLen) / (maxTotalLen - minTotalLen);

                q1 = clamp(q1, 0.0, 1.0);
                q2 = clamp(q2, 0.0, 1.0);

                float t1 = tMin * (1.0 - q1) + tMax * q1;
                float t2 = tMin * (1.0 - q2) + tMax * q2;

                if (t1 > t2)
                {
                    swap(t1, t2);
                }

                vec3 throughput = getThroughput(origin, dir, t1, t2, v_UnitCubeToCamera, phase);
                gl_FragColor = vec4(throughput * v_Color, 1.0);
            }
            return;
        }

        if(v_DistWidth > 0.0)
        {
            // still rendering "photon volume"
            float tTransMin = -1.0;
            float tTransMax = -1.0;

            bool hitMin = rayPlaneIntersect(origin, dir, v_CubeMinPlanePoint, v_CubeMinPlaneNormal, tTransMin);
            bool hitMax = rayPlaneIntersect(origin, dir, v_CubeMaxPlanePoint, v_CubeMaxPlaneNormal, tTransMax);

            if(tTransMin > tTransMax)
            {
                swap(tTransMin, tTransMax);
            }

            if(hitMin && hitMax)
            {
                float oMin = -1.0;
                float oMax = -1.0;
                if(rangeIntersect(tMin, tMax, tTransMin, tTransMax, oMin, oMax))
                {
                    tMin = oMin;
                    tMax = oMax;

                    vec3 throughput = getThroughput(origin, dir, tMin, tMax, v_UnitCubeToCamera, phase);
                    vec3 color = v_Color;
                    color *= throughput;
                    gl_FragColor = vec4(color, 1.0);

                    // NOTE: noise is in here
                    // NOTE: only for debug
                    // gl_FragColor = vec4(1.0, 0.0, 0.0, 1.0);
                }
                else
                {
                    gl_FragColor = vec4(0.0);
                }
            }
            else if(!hitMin && !hitMax)
            {
                // ray might be inside two transient planes, do a distance test
                float d0 = dot(origin - v_CubeMinPlanePoint, v_CubeMinPlaneNormal);
                float d1 = dot(origin - v_CubeMaxPlanePoint, v_CubeMaxPlaneNormal);
                // assume two transient planes have same plane normal
                if(d0 * d1 < 0.0)
                {
                    // different sign of d0 and d1 means the ray is inside two transient planes
                    vec3 throughput = getThroughput(origin, dir, tMin, tMax, v_UnitCubeToCamera, phase);
                    vec3 color = v_Color;
                    color *= throughput;
                    gl_FragColor = vec4(color, 1.0);

                    // NOTE: only for debug:
                    // gl_FragColor = vec4(0.0, 1.0, 0.0, 0.0);
                }
                else
                {
                    gl_FragColor = vec4(0.0);
                    // NOTE: only for debug:
                    // gl_FragColor = vec4(0.0, 1.0, 1.0, 0.0);
                }
            }
            else
            {
                // shouldn't happening, output magenta color for debug
                gl_FragColor = vec4(1.0, 0.0, 1.0, 1.0);
                return;
            }
        }
        else if(v_DistWidth == 0.0)
        {
            // rendering unbiased "transient photon plane"
            float tTrans = -1.0;
            bool hit = rayPlaneIntersect(origin, dir, v_CubeMinPlanePoint, v_CubeMinPlaneNormal, tTrans);
            vec4 outColor = vec4(0.0);
            if(hit)
            {
                if(tTrans >= tMin && tTrans <= tMax)
                {
                    // NOTE: assume speed of light is 1
                    const float speedOfLight = 1.0;
                    vec3 color = v_Color;

                    // NOTE: old way of calculating brightness of trasient plane
                    // float unitCubeSpaceFactor = length((v_UnitCubeToCamera * vec4(dir, 0.0)).xyz);

                    // float rayPlaneFactor = speedOfLight / (v_CubeWorldDiagLen * abs(dot(dir, v_CubeMinPlaneNormal)));
                    // color = unitCubeSpaceFactor * rayPlaneFactor * color;

                    vec3 e = normalize(v_UnitCubeToCamera[0].xyz);
                    vec3 f = normalize(v_UnitCubeToCamera[1].xyz);
                    vec3 g = normalize(v_UnitCubeToCamera[2].xyz);

                    vec3 nw = cross(e, f) + cross(f, g) + cross(g, e);
                    float dot_prod = abs(dot(nw, camRayDir.xyz));
                    // multiply determinant back
                    float det = abs(dot(e, cross(f, g)));
                    color = color * det / dot_prod;

                    if (u_EvalCamTransmittance)
                    {
                        // camera transmittance estimate
                        float camToHitPointLength = tTrans * lengthInCameraSpace(dir, v_UnitCubeToCamera);
                        vec3 camTrans = exp(-u_SigmaT * camToHitPointLength);
                        color *= camTrans * phase;
                    }

                    outColor = vec4(color, 1.0);
                }
            }
            gl_FragColor = outColor;
        }
        else
        {
            // shouldn't happening, output magenta color for debug
            gl_FragColor = vec4(1.0, 0.0, 1.0, 1.0);
            return;
        }
#else
        // non transient part code
        vec3 throughput = getThroughput(origin, dir, tMin, tMax, v_UnitCubeToCamera, phase);
        vec3 color = v_Color;
        color *= throughput;
        gl_FragColor = vec4(color, 1.0);
#endif
    }
    else
    {
        // disagreement between ray tracer and rasterizer (floating point error)
        // treat as no-hit
        gl_FragColor = vec4(0.0);
    }
}