#include "simple-common.shader"

// atrribute will receive data from a buffer
attribute vec4 a_Position;
attribute mat4 a_VolumeMatrix;
attribute vec3 a_Color;
// TRANSIENT PART
// a_DistInterval.x: minimum light ray path length for path to fall within time interval
// a_DistInterval.y: maximum light ray ...
attribute vec2 a_DistInterval;

uniform mat4 u_ModelViewProjection;
uniform mat4 u_ModelView;
uniform mat4 u_View;

varying vec4 v_ViewPosition;
// transform from the unit cube space, defined by the photon volume, to the camera space
varying mat4 v_UnitCubeToCamera;
varying mat4 v_CameraToUnitCube;
varying vec3 v_Color;

varying vec2 v_DistInterval;
varying vec3 v_LastPhotonDirection;

#ifdef ENABLE_TRANSIENT
// TRANSIENT PART
// transient planes in unit cube space
// formula for such a plane: \vec{n} \cdot \vec{p - p_0} = 0
// where p is a point on plane
varying vec3 v_CubeMinPlaneNormal;  // n
varying vec3 v_CubeMinPlanePoint;   // p
varying vec3 v_CubeMaxPlaneNormal;  // n
varying vec3 v_CubeMaxPlanePoint;   // p
varying float v_DistWidth;
varying float v_CubeWorldDiagLen;
varying vec3 v_VolumeSideLengths;
#endif

// NOTE: only works for perspective camera
// ray shooting from camera to the vertex, in camera space
// varying vec3 v_CameraRayDir;

// representation of plane:
// n * (p - p0) = 0
// n * p = s
// input: n, s
// onput: p0
vec3 findPointOnPlane(vec3 n, float s)
{
    vec3 absN = abs(n);
    if(absN.x > absN.y && absN.x > absN.z)
    {
        // choose x as dividend
        return vec3(s / n.x, 0, 0);
    }
    else if(absN.y > absN.z && absN.y > absN.x)
    {
        return vec3(0, s / n.y, 0);
    }
    else
    {
        return vec3(0, 0, s / n.z);
    }
}

void main()
{
    gl_Position = u_ModelViewProjection * a_Position;

    v_ViewPosition = u_ModelView * vec4(a_Position.xyz, 1.0);
    mat4 unitCubeToWorld = a_VolumeMatrix;
    v_UnitCubeToCamera = u_View * unitCubeToWorld;
    v_CameraToUnitCube = inverse(v_UnitCubeToCamera);

    v_Color = a_Color;
    v_DistInterval = a_DistInterval;
    v_LastPhotonDirection = normalize(unitCubeToWorld[2].xyz);

#ifdef ENABLE_TRANSIENT
    // three axis of unit cube space, written in world space
    vec3 p = unitCubeToWorld[0].xyz;
    vec3 q = unitCubeToWorld[1].xyz;
    vec3 r = unitCubeToWorld[2].xyz;
    // normal of the transient plane
    vec3 n = vec3(length(p), length(q), length(r));
    v_VolumeSideLengths = n;
    float length = length(n);
    v_CubeWorldDiagLen = length;

    v_CubeMinPlaneNormal = n / length;
    v_CubeMaxPlaneNormal = n / length;

    v_CubeMinPlanePoint = findPointOnPlane(n / length, a_DistInterval.x / length);
    v_CubeMaxPlanePoint = findPointOnPlane(n / length, a_DistInterval.y / length);

    v_DistWidth = a_DistInterval.y - a_DistInterval.x;
#endif

    // vertex position in camera space
    // vec3 eyeVertPos = (u_ModelView * a_Position).xyz;

    // NOTE: only works for perspective camera
    // direction == pos because camera is at (0, 0, 0) in camera space
    // v_CameraRayDir = normalize(eyeVertPos);
}